(*

fsharpc Mini.fs Algorithmics/Tree.fs
mono Tree.exe

fsharpi Mini.fs Algorithmics/Tree.fs

open Mini ;;
open Algorithmics.Tree ;;

*)

module Algorithmics.Tree

open Mini
open System.Collections

// Hmmm, there are generic and non-generic Enumerables, both have to
// be defined. Ugly.
type Tree<'a> =
  | Leaf
  | Node of Tree<'a> * 'a * Tree<'a>
  interface Generic.IEnumerable<'a> with
    member self.GetEnumerator () =
      let rec inorder = function
        | Leaf           -> Seq.empty
        | Node (l, x, r) -> seq { yield! inorder l; yield x; yield! inorder r}
      in (inorder self).GetEnumerator ()
  interface IEnumerable with
    member self.GetEnumerator () =
      (self :> Generic.IEnumerable<'a>).GetEnumerator() :> IEnumerator

//type List<'a> = 'a list

let rec height = function
  | Leaf           -> 0N
  | Node (l, x, r) -> 1N + max (height l) (height r)

// Suche in einem binären Suchbaum.

type Ordering = | LT | EQ | GT

let lookup (oracle : 'a -> Ordering) : Tree<'a> -> Option<'a> =
  let rec search = function
    | Leaf ->  None
    | Node (left, label, right) ->
        match oracle label with
        | LT -> search left
        | EQ -> Some label
        | GT -> search right
  in search

let rec split (n : int, list : List<'a>) : List<'a> * List<'a> =
  if n = 0
  then ([], list)
  else match list with
       | []      -> ([], [])
       | x :: xs -> let (xs1, xs2) = split (n - 1, xs)
                    (x :: xs1, xs2)

// Konstruktion eines Binärbaums aus einer Liste.

// linear-logarithmische Laufzeit:

let rec balancedTree (list : List<'a>) : Tree<'a> =
  let n = list.Length 
  if n = 0  then  Leaf
            else  let (xs1, x :: xs2) = split (n / 2, list)
                  Node (balancedTree xs1, x, balancedTree xs2)

// lineare Laufzeit:

let rec braunOfSize n xs =
    if n = 0 then (Leaf, xs)
             else let m = (n - 1) / 2
                  let (l, x :: ys) = braunOfSize m           xs
                  let (r, zs)      = braunOfSize (n - 1 - m) ys
                  (Node (l, x, r),  zs)

let braun x = fst (braunOfSize (List.length x) x)

// Überführung eines Binärbaums in eine Liste.

// quadratische Laufzeit:
let rec slowInorder (tree : Tree<'a>) : List<'a> =
  match tree with
  | Leaf                  -> []
  | Node (left, x, right) -> slowInorder left @ x :: slowInorder right

// lineare Laufzeit:

let rec inorderAppend (tree : Tree<'a>, list : List<'a>) : List<'a> =
  match tree with
  | Leaf                  -> list
  | Node (left, x, right) -> inorderAppend (left, x :: inorderAppend (right, list))

let inorder (tree : Tree<'a>) : List<'a> =
  inorderAppend (tree, [])

// Testfunktionen.

let rec leftSkewed (n : Int, x : 'a) : Tree<'a> =
  if n = 0 then Leaf else Node (leftSkewed (n - 1, x), x, Leaf)

let rec rightSkewed (n : Int, x : 'a) : Tree<'a> =
  if n = 0 then Leaf else Node (Leaf, x, rightSkewed (n - 1, x))

// slowInorder (leftSkewed (50000, 4711)) ;;
// inorder (leftSkewed (50000, 4711)) ;;

let test (generate, n) =
//  List.length (slowInorder (generate (n, "Hello, world!")))
  List.length (inorder (generate (n, "Hello, world!")))

// test (leftSkewed, 1000)
// test (rightSkewed, 1000)

// Aufteilung in den Baum ohne das letzte Element und das letzte Element selbst
let rec splitMax = function
  | Leaf           -> None
  | Node (l, a, r) ->
      Some (match splitMax r with
            | None           -> (l, a)
            | Some (r', max) -> (Node (l, a, r'), max))

// Konkatenation von Binärbäumen

let join l r =
  match splitMax l with
  | None           -> r
  | Some (l', max) -> Node (l', max, r)

// Inorderdurchlauf mit Sequenzen

module Seq =
    let rec inorder (tree : Tree<'a>) : 'a seq =
        match tree with
            | Leaf                  -> Seq.empty
            | Node (left, x, right) -> seq { yield! inorder left; yield x; yield! inorder right }

let main =
//    print (test (leftSkewed, 50000))
    print (test (rightSkewed, 50000))
