(*
Tail recursion and continuation-passing style.

fsharpc Mini.fs Algorithmics/Tree.fs Effects/Tree.fs
mono Tree.exe

fsharpi Mini.fs Algorithmics/Tree.fs Effects/Tree.fs
open Algorithmics.Tree ;;
open Effects.Tree ;;
*)

module Effects.Tree

open Mini
open Algorithmics.Tree

(*
type Tree<'elem> =
  | Leaf
  | Node of Tree<'elem> * 'elem * Tree<'elem>
*)

let leftSkewed size (x : 'elem) =
  let rec bottomUp n (tree : Tree<'elem>) =
    if n = 0 then tree else bottomUp (n - 1) (Node (tree, x, Leaf))
  bottomUp size Leaf

let rec rightSkewed size (x : 'elem) =
  let rec bottomUp n (tree : Tree<'elem>) =
    if n = 0 then tree else bottomUp (n - 1) (Node (Leaf, x, tree))
  bottomUp size Leaf

// running example

let rec product tree =
  match tree with
  | Leaf           -> 1
  | Node (l, x, r) -> product l * x * product r

// stack overflow:
// product (leftSkewed 1000000 1) ;;
// product (rightSkewed 1000000 1) ;;

///////////////////////////////////////////////////////////////////////////////
// accumulating argument
///////////////////////////////////////////////////////////////////////////////

let rec productAcc acc tree =
  match tree with
  | Leaf           -> acc
  | Node (l, x, r) -> productAcc (productAcc acc l * x) r

// stack overflow:
// productAcc 1 (leftSkewed 1000000 1) ;;
// works:
// productAcc 1 (rightSkewed 1000000 1) ;;

// explicit stack

let rec productMany acc = function
  | []                   -> acc
  | Leaf :: ts           -> productMany acc ts
  | Node (l, x, r) :: ts -> productMany (acc * x) (l :: r :: ts)

// works:
// productMany 1 [leftSkewed 1000000 1] ;;
// productMany 1 [rightSkewed 1000000 1] ;;

///////////////////////////////////////////////////////////////////////////////
// continuation passing style
///////////////////////////////////////////////////////////////////////////////

let rec productCont tree cont =
  match tree with
  | Leaf           -> cont 1
  | Node (l, x, r) -> productCont l (fun pl -> productCont r (fun pr -> cont (pl * x * pr)))

// works:
// productCont (leftSkewed 1000000 1) (fun n -> n) ;;
// productCont (rightSkewed 1000000 1) (fun n -> n) ;;

// accumulating argument and continuation passing style

let rec productAccCont acc tree cont =
  match tree with
  | Leaf           -> cont acc
  | Node (l, x, r) -> productAccCont acc l (fun pl -> productAccCont (pl * x) r cont)

// works:
// productAccCont 1 (leftSkewed 1000000 1) (fun n -> n) ;;
// productAccCont 1 (rightSkewed 1000000 1) (fun n -> n) ;;

let test () =
   print (productAccCont 1 (leftSkewed 100000000 1) (fun n -> n))
   print (productCont (leftSkewed 100000000 1) (fun n -> n))

#if INTERACTIVE
#else
let main =
    test ()
#endif
