module Grammar.AcceptorExpr

open Mini
open Grammar.Token

(*

fsharpc --nowarn:62 Mini.fs Grammar/Token.fs Grammar/AcceptorExpr.fs

Akzeptoren für einfache arithmetische Ausdrücke:

  rec expr0 -> expr1 | expr1 + expr0
  and expr1 -> expr2 | expr2 * expr1
  and expr2 -> nat | ( expr0 )

Linksfaktorisierte Variante:

  rec expr0 -> expr1 (eps | + expr0)
  and expr1 -> expr2 (eps | * expr1)
  and expr2 -> nat | ( expr0 )

*)

type Follow = List<Token> -> Bool

type Acceptor = Follow -> Follow

let symbol (a : Token) : Acceptor =
  fun (follow : Follow) ->
    fun input ->
      match input with
      | []         ->  false
      | b :: rest  ->  a = b && follow rest

let end_of_input =
  fun input -> 
    match input with
    | []      ->  true
    | _ :: _  ->  false

///////////////////////////////////////////////////////////////////////////////

module AdHoc =
  // Ad-hoc Akzeptor (linksfaktorisiert).

  let accept_RParen (follow : Follow) : Follow =
    fun input ->
      match input with
      | RParen :: rest -> follow rest
      | _              -> false

  let rec accept_expr0 (follow : Follow) : Follow =
        accept_expr1 (
          fun input ->
            match input with 
            | Plus :: rest -> accept_expr0 follow rest
            | _            -> follow input)

     and     accept_expr1 (follow : Follow) : Follow =
        accept_expr2 (
          fun input ->
            match input with
            | Asterisk :: rest -> accept_expr1 follow rest
            | _                -> follow input)

     and     accept_expr2 (follow : Follow) : Follow =
        fun input ->
          match input with
          | Num n :: rest  -> follow rest
          | LParen :: rest -> accept_expr0 (accept_RParen follow) rest
          | _              -> false

  let well_formed (input : String) : Bool =
    accept_expr0 end_of_input (lex (explode input))

  // open Grammar.AcceptorExpr.AdHoc
  // well_formed "4711+815*2765"

///////////////////////////////////////////////////////////////////////////////

module Systematic =

  // Akzeptor mit System (Original-Grammatik).

  let accept_num (follow : Follow) : Follow =
    fun input ->
        match input with
        | Num _ :: rest -> follow rest
        | _             -> false

  let rec accept_expr0 (follow : Follow) : Follow =
    fun input ->
         accept_expr1 follow input
      || accept_expr1 (symbol Plus (accept_expr0 follow)) input

  and      accept_expr1 (follow : Follow) : Follow =
    fun input ->
         accept_expr2 follow input
      || accept_expr2 (symbol Asterisk (accept_expr1 follow)) input

  and      accept_expr2 (follow : Follow) : Follow =
    fun input ->
         accept_num follow input
      || symbol LParen (accept_expr0 (symbol RParen follow)) input

  let well_formed (input : String) : Bool =
    accept_expr0 end_of_input (lex (explode input))

  // open Grammar.AcceptorExpr.Systematic
  // well_formed "4711+815*2765"

///////////////////////////////////////////////////////////////////////////////
