(*

fsharpc --nowarn:62 Mini.fs Grammar/Alphabet.fs Grammar/Token.fs Grammar/GenericAcceptor.fs

*)
module Grammar.GenericAcceptor

open Mini

type Follow<'a> = List<'a> -> Bool

type Acceptor<'a> = Follow<'a> -> Follow<'a>

let symbol (a : 'a) (follow : Follow<'a>) : Follow<'a> = fun input ->
    match input with
    | []        -> putline ("symbol " ^ show a ^ " fails"); false
    | b :: rest -> if a = b then putline ("symbol " ^ show a ^ " succeeds"); follow rest 
                            else putline ("symbol " ^ show a ^ " fails"); false

let epsilon : Acceptor<'a> =
    fun (follow : Follow<'a>) ->
        follow

let seq (accept1 : Acceptor<'a>, accept2 : Acceptor<'a>) : Acceptor<'a> =
    fun (follow : Follow<'a>) ->
        accept1 (accept2 follow)

let seq3 (accept1 : Acceptor<'a>, accept2 : Acceptor<'a>, accept3 : Acceptor<'a>) : Acceptor<'a> =
    seq (seq (accept1, accept2), accept3)

let empty : Acceptor<'a> =
    fun (follow : Follow<'a>) ->
        fun input -> false

let alt (accept1 : Acceptor<'a>, accept2 : Acceptor<'a>) : Acceptor<'a> =
    fun (follow : Follow<'a>) ->
        fun input ->
            accept1 follow input || accept2 follow input

let fixedPoint (functional : ('a -> 'b) -> ('a -> 'b)) : 'a -> 'b =
    let rec f (x : 'a) : 'b = functional f x in f

(*

let naive-fixed-point (f : 'a -> 'a) : 'a =
    let rec x : a = f x in x
*)

let end_of_input : Follow<'a> = fun input ->
    match input with
    | []      -> putline ("end-of-input succeeds"); true
    | _ :: _  -> putline ("end-of-input fails"); false

///////////////////////////////////////////////////////////////////////////////
// Anwendung: Klammersprache
///////////////////////////////////////////////////////////////////////////////

module Dyck =

    open Grammar.Alphabet

// RH: second alternative is better
    let rec acceptDyck (follow : Follow<Alphabet>) : Follow<Alphabet> =
//        alt (epsilon, seq (seq (symbol A, acceptDyck), symbol B)) follow
        alt (seq (seq (symbol A, acceptDyck), symbol B), epsilon) follow

    let acceptDyck' =
        fixedPoint (fun accepte ->
            alt (epsilon, seq (seq (symbol A, accepte), symbol B)))

    let test (s : String) : Bool =
        acceptDyck end_of_input (Alphabet.lex (explode s))

  // open Grammar.GenericAcceptor.Dyck
  // test "AAAAABBBBB"

(*        
  // rec x -> eps | x a x b
  function brackets (follow : Follow) : Follow = // loops on input ABAB
    alt (epsilon,
//         seq( seq (seq (brackets, symbol A), brackets), symbol B)) follow  // left recursive
         seq( seq (seq (symbol A, brackets), symbol B), brackets)) follow  // right recursive

  // rec x -> eps | a x a | b x b
  function mirror (follow : Follow) : Follow =
    alt (alt (epsilon,
              seq (seq (symbol A, mirror), symbol A)),
              seq (seq (symbol B, mirror), symbol B)) follow
  // mirror end-of-input (Alphabet.lex (explode "ABBA"))
  // quadratische Anzahl von Schritten: mirror end-of-input (Alphabet.lex (explode "A^nA^n"))
*)

///////////////////////////////////////////////////////////////////////////////
// Anwendung: einfache arithmetische Ausdrücke
///////////////////////////////////////////////////////////////////////////////

open Grammar.Token

// parser
let accept_num (follow : Follow<Token>) : Follow<Token> =
    fun input ->
      match input with
      |  Num _ :: rest -> follow rest
      | _              -> false

module Expr =
  
(*
  val accept-expr =
    naive-fixed-point (fun (accept-expr0 = accept-expr0, accept-expr1 = accept-expr1, accept-expr2 = accept-expr2) ->
      ( accept-expr0 = fun follow -> seq (accept-expr1, alt (epsilon, seq (symbol Plus, accept-expr0)))     follow,
        accept-expr1 = fun follow -> seq (accept-expr2, alt (epsilon, seq (symbol Asterisk, accept-expr1))) follow,
        accept-expr2 = fun follow -> alt (accept-num, seq3 (symbol LParen, accept-expr0, symbol RParen)) follow))
*)

  // ohne Linksfaktorisierung
  let rec accept_expr0 : Acceptor<Token> = fun follow ->
    alt (accept_expr1, seq3 (accept_expr1, symbol Plus, accept_expr0)) follow
  and      accept_expr1 : Acceptor<Token> = fun follow ->
    putline "expr1";
    alt (accept_expr2, seq3 (accept_expr2, symbol Asterisk, accept_expr1)) follow
  and      accept_expr2 : Acceptor<Token> = fun follow ->
    alt (accept_num, seq3 (symbol LParen, accept_expr0, symbol RParen)) follow

module OptExpr =

  // mit Linksfaktorisierung
  let rec accept_expr0 : Acceptor<Token> = fun follow ->
    seq (accept_expr1, alt (epsilon, seq (symbol Plus, accept_expr0))) follow
  and      accept_expr1 : Acceptor<Token> = fun follow ->
    putline "expr1";
    seq (accept_expr2, alt (epsilon, seq (symbol Asterisk, accept_expr1))) follow
  and      accept_expr2 : Acceptor<Token> = fun follow ->
    alt (accept_num, seq3 (symbol LParen, accept_expr0, symbol RParen)) follow

  let well_formed (input : String) : Bool =
    accept_expr0 end_of_input (Token.lex (explode input))

  // open Grammar.GenericAcceptor.Expr
  // well_formed "4711+815*2765"
