(*

fsharpc Mini.fs Grammar/Token.fs Grammar/Expr.fs Grammar/ParserExpr.fs

Parser für einfache arithmetische Ausdrücke:

  rec expr0 -> expr1 | expr1 + expr0
  and expr1 -> expr2 | expr2 * expr1
  and expr2 -> nat | ( expr0 )

Linksfaktorisierte Variante:

  rec expr0 -> expr1 (eps | + expr0)
  and expr1 -> expr2 (eps | * expr1)
  and expr2 -> nat | ( expr0 )

*)

module Grammar.ParserExpr

open Mini
open Grammar.Token
open Grammar.Expr

type Follow<'v> = List<Token> -> Option<'v>

// rank-2 type: not directly expressible in F#, encode using interface
// with a single polymorphic method
// type Parser<a> = <result> -> (a -> Follow<result>) -> Follow<result>

let symbol (a : Token) : Follow<Expr> -> Follow<Expr> =
  fun (follow : Follow<Expr>) ->
    fun input ->
       match input with
       | []        -> None
       | b :: rest -> if a = b then follow rest else None

let end_of_input (e : Expr) : Follow<Expr> =
  fun input ->
    match input with
    | []     -> Some e
    | _ :: _ -> None

///////////////////////////////////////////////////////////////////////////////

module AdHoc =
  // Ad-hoc Parser (linksfaktorisiert).

  let parse_RParen (follow : Follow<Expr>) : Follow<Expr> = fun input ->
    match input with
    |  RParen :: rest -> follow rest
    | _               -> None

  let rec parse_expr0 (follow : Expr -> Follow<Expr>) : Follow<Expr> =
        parse_expr1 (fun e1 ->
          fun input ->
            match input with
            | Plus :: rest -> parse_expr0 (fun e2 -> follow (Add (e1, e2))) rest
            | _            -> follow e1 input)

      and parse_expr1 (follow : Expr -> Follow<Expr>) : Follow<Expr> =
        parse_expr2 (fun e1 ->
          fun input ->
            match input with
            | Asterisk :: rest -> parse_expr1 (fun e2 -> follow (Mul (e1, e2))) rest
            | _                -> follow e1 input)

      and parse_expr2 (follow : Expr -> Follow<Expr>) : Follow<Expr> =
        fun input ->
          match input with
          | Num n :: rest  -> follow (Const n) rest
          | LParen :: rest -> parse_expr0 (fun e -> parse_RParen (follow e)) rest
          | _              -> None

  let abstract_syntax_tree (input : String) : Option<Expr> =
    parse_expr0 end_of_input (Token.lex (explode input))

  // open Grammar.ParserExpr.AdHoc
  // abstract_syntax_tree "4711+815*2765"

///////////////////////////////////////////////////////////////////////////////

module Systematic =

  // Parser mit System (Original-Grammatik).

  let parse_num (follow : Nat -> Follow<Expr>) : Follow<Expr> =
    fun input ->
      match input with
      | Num n :: rest -> follow n rest
      | _             -> None

  let rec parse_expr0 (follow : Expr -> Follow<Expr>) : Follow<Expr> =
        fun input ->
          match parse_expr1 (fun e -> follow e) input with
          | None ->
              parse_expr1 (fun e1 -> 
               symbol Plus (
                parse_expr0 (fun e2 -> 
                 follow (Add (e1, e2))))) input
          | Some e -> Some e

      and parse_expr1 (follow : Expr -> Follow<Expr>) : Follow<Expr> =
        fun input ->
          match parse_expr2 (fun e -> follow e) input with
          | None ->
              parse_expr2     (fun e1 ->
               symbol Asterisk (
                parse_expr1     (fun e2 -> 
                 follow (Mul (e1, e2))))) input
          | Some e -> Some e

      and parse_expr2 (follow : Expr -> Follow<Expr>) : Follow<Expr> =
        fun input ->
          match parse_num (fun n -> follow (Const n)) input with
          | None ->
              symbol LParen (
               parse_expr0   (fun e ->
                symbol RParen (
                 follow e))) input
          | Some e -> Some e

  let abstract_syntax_tree (input : String) : Option<Expr> =
    parse_expr0 end_of_input (Token.lex (explode input))

  // open Grammar.ParserExpr.Systematic
  // abstract_syntax_tree "4711+815*2765"

///////////////////////////////////////////////////////////////////////////////
