(*

Natürliche Zahlen, Wahrheitswerte, arithmetische und Boolesche
Ausdrücke.

fsharpi Mini.fs Interpreter/Booleans.fs

open Interpreter.Booleans ;;

*)

module Interpreter.Booleans

// abstrakte Syntax

type Op =
    | Add | Sub | Mul | Div | Mod
    | Lt  | Lte | Equ | Neq | Gte | Gt

type Expr =
    | False
    | True
    | If  of Expr * Expr * Expr
    | Num of Nat
    | Bin of Expr * Op * Expr

// syntaktischer Zucker
let add (e1, e2) = Bin (e1, Add, e2)    
let sub (e1, e2) = Bin (e1, Sub, e2)    
let mul (e1, e2) = Bin (e1, Mul, e2)    
let equ (e1, e2) = Bin (e1, Equ, e2)    

// dynamische Semantik

type Value =
    | NAT  of Nat
    | BOOL of Bool

let primitive : Value * Op * Value -> Value option = function
    | (NAT n1, Add, NAT n2) -> Some (NAT  (n1 + n2))
    | (NAT n1, Sub, NAT n2) -> Some (NAT  (n1 - n2))
    | (NAT n1, Mul, NAT n2) -> Some (NAT  (n1 * n2))
    | (NAT n1, Div, NAT n2) -> Some (NAT  (n1 / n2))
    | (NAT n1, Mod, NAT n2) -> Some (NAT  (n1 % n2))
    | (NAT n1, Lt,  NAT n2) -> Some (BOOL (n1 < n2))
    | (NAT n1, Lte, NAT n2) -> Some (BOOL (n1 <= n2))
    | (NAT n1, Equ, NAT n2) -> Some (BOOL (n1 = n2))
    | (NAT n1, Neq, NAT n2) -> Some (BOOL (n1 <> n2))
    | (NAT n1, Gte, NAT n2) -> Some (BOOL (n1 >= n2))
    | (NAT n1, Gt,  NAT n2) -> Some (BOOL (n1 > n2))
    | _                     -> None

type Frame =
    | If1  of Expr * Expr
    | Bin1 of Op * Expr
    | Bin2 of Value * Op

type State =
    | Eval of Expr  * Frame list
    | Ret  of Value * Frame list

let step : State -> State = function
// Auswertung
    | Eval (False,            stack)        -> Ret (BOOL false, stack)
    | Eval (True,             stack)        -> Ret (BOOL true,  stack)
    | Eval (If (e1, e2, e3),  stack)        -> Eval (e1, If1 (e2, e3) :: stack) 
    | Eval (Num n,            stack)        -> Ret (NAT n, stack)
    | Eval (Bin (e1, op, e2), stack)        -> Eval (e1, Bin1 (op, e2) :: stack)
// Rückgabe
    | Ret  (v,      [])                     -> Ret  (v, [])
    | Ret  (BOOL b, If1 (e2, e3)  :: stack) -> if b then Eval (e2, stack)
                                                    else Eval (e3, stack)
    | Ret  (_,      If1 (e2, e3)  :: stack) -> error "type mismatch"
    | Ret  (v1,     Bin1 (op, e2) :: stack) -> Eval (e2, Bin2 (v1, op) :: stack)
    | Ret  (v2,     Bin2 (v1, op) :: stack) -> match primitive (v1, op, v2) with
                                               | Some v -> Ret (v, stack)
                                               | None   -> error "type mismatch"
let eval (e : Expr) : Value =
    let rec loop = function
        | Ret  (v, []) -> v
        | state        -> loop (step state)
    loop (Eval (e, []))

// Beispiele

let ex1 = add (Num 4N, Num 7N)
let ex2 = add (Num 1N, Num 2N)
let ex3 = add (ex1, ex2)

// Eval (ex3, []) ;;
// step it ;;

let rec leftist n =
    if n = 0N then Num 0N
              else add (leftist (n - 1N), Num n)

let rec rightist n =
    if n = 0N then Num 0N
              else add (Num n, rightist (n - 1N))

// Eval (leftist  5N, []) ;;
// Eval (rightist 5N, []) ;;
// step it ;;
// eval (leftist  1000N) ;;
// eval (rightist 1000N) ;;
