(*

Natürliche Zahlen, Wahrheitswerte, arithmetische Ausdrücke, Boolesche
Ausdrücke und Wertedefinitionen.

fsharpi Mini.fs Interpreter/Definitions.fs

open Interpreter.Definitions ;;

*)

module Interpreter.Definitions
#nowarn "62"

// endliche Abbildungen

type Map<'key, 'value> =
    | Empty
    | Comma of Map<'key, 'value> * 'key * 'value

let rec lookup (x : 'key) : Map<'key, 'value> -> 'value option  when 'key : equality = function
    | Empty               -> None
    | Comma (δ, x1, v1) -> if x = x1 then Some v1 else lookup x δ

// abstrakte Syntax

type Op =
    | Add | Sub | Mul | Div | Mod
    | Lt  | Lte | Equ | Neq | Gte | Gt

type Ident =
    String

type Expr =
    | Num of Nat
    | Bin of Expr * Op * Expr
    | False
    | True
    | If  of Expr * Expr * Expr
    | Id  of Ident
    | Let of Ident * Expr * Expr

// syntaktischer Zucker
let add (e1, e2) = Bin (e1, Add, e2)    
let sub (e1, e2) = Bin (e1, Sub, e2)    
let mul (e1, e2) = Bin (e1, Mul, e2)    
let equ (e1, e2) = Bin (e1, Equ, e2)    

let rec local = function
    | ([],             e) -> e
    | ((x1, e1) :: ds, e) -> Let (x1, e1, local (ds, e))

// dynamische Semantik

type Value =
    | BOOL of Bool
    | NAT  of Nat

let primitive : Value * Op * Value -> Value option = function
    | (NAT n1, Add, NAT n2) -> Some (NAT  (n1 + n2))
    | (NAT n1, Sub, NAT n2) -> Some (NAT  (n1 - n2))
    | (NAT n1, Mul, NAT n2) -> Some (NAT  (n1 * n2))
    | (NAT n1, Div, NAT n2) -> Some (NAT  (n1 / n2))
    | (NAT n1, Mod, NAT n2) -> Some (NAT  (n1 % n2))
    | (NAT n1, Lt,  NAT n2) -> Some (BOOL (n1 < n2))
    | (NAT n1, Lte, NAT n2) -> Some (BOOL (n1 <= n2))
    | (NAT n1, Equ, NAT n2) -> Some (BOOL (n1 = n2))
    | (NAT n1, Neq, NAT n2) -> Some (BOOL (n1 <> n2))
    | (NAT n1, Gte, NAT n2) -> Some (BOOL (n1 >= n2))
    | (NAT n1, Gt,  NAT n2) -> Some (BOOL (n1 > n2))
    | _                     -> None

type Env = Map<Ident, Value>

type Frame =
    | Bin1 of Env * Op * Expr
    | Bin2 of       Value * Op
    | If1  of Env * Expr * Expr
    | Let1 of Env * Ident * Expr

type State =
    | Ret  of Value      * Frame list
    | Eval of Env * Expr * Frame list

let step : State -> State = function
// Auswertung
    | Eval (δ, Num n,            stack) -> Ret (NAT n, stack)
    | Eval (δ, Bin (e1, op, e2), stack) -> Eval (δ, e1, Bin1 (δ, op, e2) :: stack)
    | Eval (δ, False,            stack) -> Ret (BOOL false, stack)
    | Eval (δ, True,             stack) -> Ret (BOOL true,  stack)
    | Eval (δ, If (e1, e2, e3),  stack) -> Eval (δ, e1, If1 (δ, e2, e3) :: stack) 
    | Eval (δ, Id x,             stack) -> match lookup x δ with
                                             | None   -> error ("'" ^ x ^ "' is undefined")
                                             | Some v -> Ret (v, stack) 
    | Eval (δ, Let (x1, e1, e),  stack) -> Eval (δ, e1, Let1 (δ, x1, e) :: stack)
// Rückgabe
    | Ret  (v,      [])                          -> Ret  (v, [])
    | Ret  (v1,     Bin1 (δ, op, e2) :: stack) -> Eval (δ, e2, Bin2 (v1, op) :: stack)
    | Ret  (v2,     Bin2 (v1, op)      :: stack) -> match primitive (v1, op, v2) with
                                                    | Some v -> Ret (v, stack)
                                                    | None   -> error "type mismatch"
    | Ret  (BOOL b, If1 (δ, e2, e3)  :: stack) -> if b then Eval (δ, e2, stack)
                                                    else Eval (δ, e3, stack)
    | Ret  (_,      If1 (δ, e2, e3)  :: stack) -> error "type mismatch"
    | Ret  (v1,     Let1 (δ, x1, e)  :: stack) -> Eval (Comma (δ, x1, v1), e, stack)

let eval (e : Expr) : Value =
    let rec loop = function
        | Ret  (v, []) -> v
        | state        -> loop (step state)
    loop (Eval (Empty, e, []))

// Beispiele

let ex1 = add (Num 1N, Num 2N)
let ex2 = add (Num 3N, Num 4N)
let ex3 = add (ex1, ex2)

// Eval (ex3, []) ;;
// step it ;;

let rec leftist n =
    if n = 0N then Num 0N
              else add (leftist (n - 1N), Num n)

let rec rightist n =
    if n = 0N then Num 0N
              else add (Num n, rightist (n - 1N))

// Eval (leftist  5N, []) ;;
// Eval (rightist 5N, []) ;;
// step it ;;
// steps (Eval (leftist  1000N, [])) ;;
// steps (Eval (rightist 1000N, [])) ;;

let ex4 = local ([("a", Num 47N);
                  ("b", add (Id "a", Num 11N))],
                 mul (Id "a", Id "b"))

// Eval (Empty, ex4, []) ;;
