(*

Natürliche Zahlen, Wahrheitswerte, arithmetische Ausdrücke, Boolesche
Ausdrücke, Wertedefinitionen und (rekursive) Funktionen.

fsharpi Mini.fs Interpreter/Functions.fs

open Interpreter.Functions

*)

module Interpreter
#nowarn "62"

// endliche Abbildungen

type Map<'key, 'value> =
    | Empty
    | Comma of Map<'key, 'value> * 'key * 'value

let rec lookup x = function
    | Empty               -> None
    | Comma (δ, x1, v1) -> if x = x1 then Some v1 else lookup x δ

// abstrakte Syntax

type Op =
    | Add | Sub | Mul | Div | Mod
    | Lt  | Lte | Equ | Neq | Gte | Gt

type Ident =
    String

type Expr =
    | Num of Nat
    | Bin of Expr * Op * Expr
    | False
    | True
    | If  of Expr * Expr * Expr
    | Id  of Ident
    | Let of Ident * Expr * Expr
    | Fun    of Ident * Expr
    | App    of Expr * Expr
    | Letrec of Ident * Ident * Expr * Expr

// syntaktischer Zucker
let add (e1, e2) = Bin (e1, Add, e2)    
let sub (e1, e2) = Bin (e1, Sub, e2)    
let mul (e1, e2) = Bin (e1, Mul, e2)    
let equ (e1, e2) = Bin (e1, Equ, e2)
// let (^*^) e1 e2 = Bin (e1, Mul, e2)    

let rec local = function
    | ([],             e) -> e
    | ((x1, e1) :: ds, e) -> Let (x1, e1, local (ds, e))

let rec funs = function
    | ([],      e) -> e
    | (x :: xs, e) -> Fun (x, funs (xs, e))
let rec apps f = function
    | []      -> f
    | e :: es -> apps (App (f, e)) es

// dynamische Semantik

type Value =
    | BOOL of Bool
    | NAT  of Nat
    | Closure    of Env * Ident * Expr 
//    | Recclosure of Env * Ident * Ident * Expr 
and Env = Map<Ident, Value>

let primitive : Value * Op * Value -> Value option = function
    | (NAT n1, Add, NAT n2) -> Some (NAT  (n1 + n2))
    | (NAT n1, Sub, NAT n2) -> Some (NAT  (n1 - n2))
    | (NAT n1, Mul, NAT n2) -> Some (NAT  (n1 * n2))
    | (NAT n1, Div, NAT n2) -> Some (NAT  (n1 / n2))
    | (NAT n1, Mod, NAT n2) -> Some (NAT  (n1 % n2))
    | (NAT n1, Lt,  NAT n2) -> Some (BOOL (n1 < n2))
    | (NAT n1, Lte, NAT n2) -> Some (BOOL (n1 <= n2))
    | (NAT n1, Equ, NAT n2) -> Some (BOOL (n1 = n2))
    | (NAT n1, Neq, NAT n2) -> Some (BOOL (n1 <> n2))
    | (NAT n1, Gte, NAT n2) -> Some (BOOL (n1 >= n2))
    | (NAT n1, Gt,  NAT n2) -> Some (BOOL (n1 > n2))
    | _                     -> None

type Frame =
    | Bin1 of Env * Op * Expr
    | Bin2 of       Value * Op
    | If1  of Env * Expr * Expr
    | Let1 of Env * Ident * Expr
    | App1 of Env * Expr
//    | App2 of       Value

type State =
    | Ret  of Value      * Frame list
    | Eval of Env * Expr * Frame list

let step : State -> State = function
// Auswertung
    | Eval (δ, Num n,            stack) -> Ret (NAT n, stack)
    | Eval (δ, Bin (e1, op, e2), stack) -> Eval (δ, e1, Bin1 (δ, op, e2) :: stack)
    | Eval (δ, False,            stack) -> Ret (BOOL false, stack)
    | Eval (δ, True,             stack) -> Ret (BOOL true,  stack)
    | Eval (δ, If (e1, e2, e3),  stack) -> Eval (δ, e1, If1 (δ, e2, e3) :: stack) 
    | Eval (δ, Id x,             stack) -> match lookup x δ with
                                             | None   -> error ("'" ^ x ^ "' is undefined")
                                             | Some v -> Ret (v, stack) 
    | Eval (δ, Let (x1, e1, e),  stack) -> Eval (δ, e1, Let1 (δ, x1, e) :: stack)
    | Eval (δ, Fun (x, e),       stack) -> Ret (Closure (δ, x, e), stack)
    | Eval (δ, App (e, e1),      stack) -> Eval (δ, e, App1 (δ, e1) :: stack)
//    | Eval (δ, Letrec (f, x1, e, e'), stack) -> Eval (Comma (δ, f, Recclosure (δ, f, x1, e)), e', stack)
    | Eval (δ, Letrec (f, x1, e, e'), stack) ->
        let rec δ' = Comma (δ, f, c)
        and     c    = Closure (δ', x1, e)
        in Eval (δ', e', stack)
// Rückgabe
    | Ret  (v,      [])                          -> Ret  (v, [])
    | Ret  (v1,     Bin1 (δ, op, e2) :: stack) -> Eval (δ, e2, Bin2 (v1, op) :: stack)
    | Ret  (v2,     Bin2 (v1, op)      :: stack) -> match primitive (v1, op, v2) with
                                                    | Some v -> Ret (v, stack)
                                                    | None   -> error "type mismatch"
    | Ret  (BOOL b, If1 (δ, e2, e3)  :: stack) -> if b then Eval (δ, e2, stack)
                                                    else Eval (δ, e3, stack)
    | Ret  (_,      If1 (δ, e2, e3)  :: stack) -> error "type mismatch"
    | Ret  (v1,     Let1 (δ, x1, e)  :: stack) -> Eval (Comma (δ, x1, v1), e, stack)

    | Ret  (Closure (δ', x1, e'), App1 (δ, e1)  :: stack) -> Eval (δ, e1, Let1 (δ', x1, e') :: stack)
    | Ret  (v,    App1 _                      :: stack) -> error "type mismatch"

(*
    | Ret  (v,     App1 (δ, e1)              :: stack) -> Eval (δ, e1, App2 v :: stack)
    | Ret  (v1,    App2 (Closure (δ, x1, e)) :: stack) -> Eval (Comma (δ, x1, v1), e, stack)
//    | Ret  (v1,    App2 (c & Recclosure (δ, f, x1, e)) :: stack) -> Eval (Comma (Comma (δ, f, c), x1, v1), e, stack)
    | Ret  (v1,    App2 _                      :: stack) -> error "type mismatch"
*)    

let rec iterate n f x =
    if n = 0 then x else f (iterate (n - 1) f x)

let eval (e : Expr) : Value =
    let rec loop = function
        | Ret  (v, []) -> v
        | state        -> loop (step state)
    loop (Eval (Empty, e, []))

let countSteps (e : Expr) : Value * Nat =
    let rec loop = function
        | (Ret  (v, []),  n) -> (v, n)
        | (state,         n) -> loop (step state, n + 1N)
    loop (Eval (Empty, e, []), 0N)

// Beispiele

let ex1 = add (Num 1N, Num 2N)
let ex2 = add (Num 3N, Num 4N)
let ex3 = add (ex1, ex2)

// Eval (ex3, []) ;;
// step it ;;

let rec leftist n =
    if n = 0N then Num 0N
              else add (leftist (n - 1N), Num n)

let rec rightist n =
    if n = 0N then Num 0N
              else add (Num n, rightist (n - 1N))

// Eval (leftist  5N, []) ;;
// Eval (rightist 5N, []) ;;
// step it ;;
// countSteps (Eval (leftist  1000N, [])) ;;
// countSteps (Eval (rightist 1000N, [])) ;;

let ex4 = local ([("a", Num 47N);
                  ("b", add (Id "a", Num 11N))],
                 mul (Id "a", Id "b"))

// Eval (Empty, ex4, []) ;;

// Laufzeitmessungen

let Twice  = funs (["f"; "x"], App (Id "f", App (Id "f", Id "x")))
let Thrice = funs (["f"; "x"], App (Id "f", App (Id "f", App (Id "f", Id "x"))))
let Succ   = Fun ("n", add (Id "n", Num 1N))
let Double = Fun ("n", mul (Id "n", Num 2N))

let twice  = fun f x -> f (f x)
let thrice = fun f x -> f (f (f x))
let succ   = fun n -> n + 1N

// Eval (Empty, App (App (Twice, Succ), Num 0N), []) ;;
// Eval (Empty, App (App (App (Twice, Twice), Succ), Num 0N), []) ;;
// Eval (Empty, App (App (App (App (Twice, Twice), Twice), Succ), Num 0N), []) ;;
// Eval (Empty, App (App (App (App (App (Twice, Twice), Twice), Twice), Succ), Num 0N), []) ;;
// Eval (Empty, App (App (App (App (App (Twice, Twice), Twice), Thrice), Succ), Num 0N), []) ;;
// eval (apps Twice [Twice; Twice; Thrice; Succ; Num 0N]) ;;

// Stack Overflows und Endrekursion

let Factorial =
    Letrec ("fac", "n", If (equ (Id "n", Num 0N), Num 1N, mul (Id "n", App (Id "fac", sub (Id "n", Num 1N)))), Id "fac")
//    Letrec ("fac", "n", If (equ (Id "n", Num 0N), Num 1N, Id "n" ^*^ App (Id "fac", sub (Id "n", Num 1N))), Id "fac")
let Sum =
    Letrec ("sum", "n", If (equ (Id "n", Num 0N), Num 0N, add (Id "n", App (Id "sum", sub (Id "n", Num 1N)))), Id "sum")

// Eval (Empty, App (Factorial, Num 10N), []) ;;
// countSteps it ;;

let rec factorial n =
    if n = 0N then 1N else n * factorial (n - 1N)
let rec sum n =
    if n = 0N then 0N else n + sum (n - 1N)

let rec accumulate a n =
    if n = 0N then a else accumulate (a + n) (n - 1N)

// #time ;;
// sum 50000N ;;
// Eval (Empty, App (Sum, Num 50000N), []) ;;
// eval it ;;

// stack overflow! für n > 50000

let Forever =
    Letrec ("forever", "n", App (Id "forever", add (Id "n", Num 1N)), Id "forever")
//    Letrec ("forever", "n", App (Id "forever", Id "n"), Id "forever")

// Eval (Empty, App (Forever, Num 0N), []) ;;
// step it ;;

// Der strikte Fixpunkt-Kombinator Z (der eta-expandierte Y-Kombinator)
// Z = λ f . (λ x . f (λ v . x x v)) (λ x . f (λ v . x x v))
let Z = let x = Fun ("x", App (Id "f", Fun ("v", App (App (Id "x", Id "x"), Id "v")))) in Fun ("f", App (x, x))

let Factorial' = App (Z, Fun ("fac", Fun ("n", If (equ (Id "n", Num 0N), Num 1N, mul (Id "n", App (Id "fac", sub (Id "n", Num 1N)))))))
// eval (App (Factorial', Num 100N)) ;;
