(*
Bank accounts using modules, interfaces and object expressions, and
classes.

fsharpc Mini.fs Objects/Bank.fs
mono Bank.exe

fsharpi Mini.fs Objects/Bank.fs
open Objects.Bank ;;
*)

module Objects.Bank
#nowarn "62"

open Mini

exception Insufficient of Nat

let monus (n1 : Nat, n2 : Nat) : Nat =
  if n1 >= n2 then
    n1 - n2
  else
    raise (Insufficient n1)

///////////////////////////////////////////////////////////////////////////////
// modules
///////////////////////////////////////////////////////////////////////////////

module Account =

  let mutable private funds = 0N

  let deposit (amount : Nat) =
      funds <- funds + amount

  let withdraw (amount : Nat) =
      funds <- monus (funds, amount)

  let balance () = funds

///////////////////////////////////////////////////////////////////////////////
// interfaces and object expressions
///////////////////////////////////////////////////////////////////////////////

//// interfaces

type IAccount =
    interface
      abstract member Deposit  : Nat -> Unit
      abstract member Withdraw : Nat -> Unit
      abstract member Balance  : Nat
    end

let transfer (acc1 : IAccount, amount : Nat, acc2 : IAccount) =
  acc1.Withdraw amount;
  acc2.Deposit  amount

//// object expressions

// single accounts

let my_account : IAccount =
  let mutable funds = 0815N
  {
    new IAccount with
      member self.Deposit (amount : Nat) =
        funds <- funds + amount

      member self.Withdraw (amount : Nat) =
        funds <- monus (funds, amount)

      member self.Balance =
        funds
  }

// banks using modules and object constructors

exception Limit

module TrustMe =  
  do putline "TrustMe is founded (module)."

  let BIC = 4711N

  let mutable private no = 0N

  let account (seed : Nat) =
    do no <- no + 1N
    let mutable funds = seed
    {
      new IAccount with
        member self.Deposit (amount : Nat) =
          funds <- funds + amount

        member self.Withdraw (amount : Nat) =
          funds <- monus (funds, amount)

        member self.Balance =
          funds
    }

  let no_of_accounts () = no

// delegation

  let limit = 1000N

  let student_account (seed : Nat) =
    let basic = account seed
    {
      new IAccount with
        member self.Deposit amount = basic.Deposit amount

        member self.Withdraw amount =
          if amount > limit then
            raise Limit
          else
            basic.Withdraw amount

        member self.Balance = basic.Balance
    }

let ralfs = TrustMe.account 4711N

// ralfs.Balance ;;
// ralfs.Deposit 0815N ;;
// ralfs.Balance ;;

//// sub-interfaces

type IAccountPlus =
    interface
      inherit IAccount
      abstract member Statement : Array<Nat>
    end

let your_account : IAccountPlus =
  let size      = 10
  let history   = [| for k in 0 .. size - 1 -> 0N |]
  let mutable i = 0
  let next ()   = (i + 1) % size
  {
    new IAccountPlus with
      member self.Deposit (amount : Nat) =
//        history.[next ()] <- history.[i] + amount; i <- next ()
        let i' = (i + 1) % size
        history.[i'] <- history.[i] + amount
        i <- i'

      member self.Withdraw (amount : Nat) =
        history.[next ()] <- monus (history.[i],  amount); i <- next ()

      member self.Balance =
        history.[i]

      member self.Statement =
        [| for k in 0 .. size - 1 -> history.[(size + i - k) % size] |]
  }

// transfer (my_account, 100N, your_account);;

///////////////////////////////////////////////////////////////////////////////
// classes
///////////////////////////////////////////////////////////////////////////////

type TrustMeC (seed : Nat) =
  class
    static do putline "TrustMe is founded (class)."  // once per class
    
    static let bgreeting = "Hiya"

    static let mutable no = 0  // class variable
    
    do no <- no + 1  // once per object
    
    let mutable funds = seed  // instance variable

    static member private BIC = 4711N

    static member no_of_accounts = no

    new () = TrustMeC 10N

    member self.Deposit (amount : Nat) =
      funds <- funds + amount

    abstract member Withdraw : Nat -> Unit
    default self.Withdraw (amount : Nat) =
      funds <- monus (funds, amount)

    member self.Balance =
      funds

    interface IAccount with
      member self.Deposit amount = self.Deposit amount  // not a recursive call
      member self.Withdraw amount = self.Withdraw amount
      member self.Balance = self.Balance
  end  

type TrustMeGold (seed : Nat) =
  class
    inherit TrustMeC (seed)

    member self.Clear () : Nat =
      let amount = self.Balance
      self.Withdraw amount
      amount
  end

type TrustMeStudent (seed : Nat, limit : Nat) =
  class
    inherit TrustMeC (seed)

    let limit = min limit 1000N

    override self.Withdraw (amount : Nat) =
      if amount > limit then
        raise Limit
      else
        base.Withdraw amount
  end

type CheapNEasy (seed : Nat) =
  class
    let mutable funds_log = [seed]

    static member BIC = 1234N

    // private to enclosing assembly (DLL or EXE); should be "protected"
    member internal self.Log = funds_log

    interface IAccount with
      member self.Deposit (amount : Nat) =
        funds_log <- (List.head funds_log + amount) :: funds_log

      member self.Withdraw (amount : Nat) =
        funds_log <- monus (List.head funds_log, amount) :: funds_log

      member self.Balance =
        List.head funds_log

    member self.Cancel =
      funds_log <- List.tail funds_log

    // expose interface
    member self.Deposit amount = (self :> IAccount).Deposit amount 
    member self.Withdraw amount = (self :> IAccount).Withdraw amount
    member self.Balance = (self :> IAccount).Balance
    // NB "self.Balance = self.Balance" is an infinite loop
  end

let lisas = CheapNEasy 100N

type CheapNEasyPlus (seed : Nat) =
  class
    inherit CheapNEasy (seed)  // don't omit the parentheses

    interface IAccountPlus with
      member self.Statement =
        [| for x in List.take 10 self.Log -> x |]

    // expose interface
    member self.Statement = (self :> IAccountPlus).Statement
  end

type OnlineBank (seed : Nat, size : Int) as self =
  class
    let size      = min size 365
    let history   = [| for k in 0 .. size - 1 -> 0N |]
    let mutable i = 0
    let next ()   = (i + 1) % size
    do (self :> IAccount).Deposit seed

    static member BIC = 0815N

    interface IAccount with
      member self.Deposit (amount : Nat) =
        history.[next ()] <- history.[i] + amount; i <- next ()

      member self.Withdraw (amount : Nat) =
        history.[next ()] <- monus (history.[i],  amount); i <- next ()

      member self.Balance =
        history.[i]

    interface IAccountPlus with
      member self.Statement =
        [| for k in 0 .. size - 1 -> history.[(size + i - k) % size] |]
  end

// let ralf = TrustMe 4711N ;;
// let harry = CheapNEasy 0N ;;
// transfer (ralf, 100N, harry) ;;
// (ralf :> IAccount).Balance;;
// (harry :> IAccount).Balance;;      
