(*
Enumerators and enumerables (from scratch, that is, redefining
prefined interfaces etc).

fsharpi Mini.fs Algorithmics/Tree.fs Objects/Enumerator.fs
open Objects.Enumerator ;;
*)

module Objects.Enumerator

open Mini
open Algorithmics.Tree

///////////////////////////////////////////////////////////////////////////////
// enumerators
///////////////////////////////////////////////////////////////////////////////

type IEnumerator<'elem> =
  interface
    abstract member Current  : 'elem
    abstract member MoveNext : Unit -> Bool
  end

let notStarted () =
    invalidOp "Enumeration has not started. Call MoveNext."
let alreadyFinished () =
    invalidOp "Enumeration already finished."

let range (lower, upper) =
    let mutable started = false 
    let mutable current = lower
    { new IEnumerator<Int> with
          member self.Current =
              if started then
                  if current <= upper then current
                                      else alreadyFinished ()
              else 
                  notStarted ()
          member self.MoveNext () =
              if started then
                  current <- current + 1
              else 
                  started <- true
              current <= upper
    }

(*
> let x = range(2, 3) ;;
val x : IEnumerator<Int>
> x.Current ;;
System.InvalidOperationException: Enumeration has not started. Call MoveNext.
> x.MoveNext () ;;
val it : Bool = true
> x.Current ;;    
val it : Int = 2
> x.MoveNext () ;;
val it : Bool = true
> x.Current ;;    
val it : Int = 3
> x.MoveNext () ;;
val it : Bool = false
> x.Current ;;    
System.InvalidOperationException: Enumeration already finished.
*)

let squares () =
    let mutable started = false 
    let mutable current = 0N
    let mutable odd = 1N
    { new IEnumerator<Nat> with
          member self.Current =
              if started then
                  current
              else 
                  notStarted ()
          member self.MoveNext () =
              if started then
                  current <- current + odd
                  odd <- odd + 2N
              else 
                  started <- true
              true
    }

let nonEmpty = function
    | [] -> false
    | _  -> true

let fromList (list : List<'elem>) : IEnumerator<'elem> =
    let mutable started = false 
    let mutable current = list
    { new IEnumerator<'elem> with
          member self.Current =
              if started then
                match current with
                | []       -> alreadyFinished ()
                | hd :: tl -> hd
              else 
                  notStarted ()
          member self.MoveNext () =
              if started then
                match current with
                | []       -> false
                | hd :: tl -> current <- tl
                              nonEmpty current
              else 
                  started <- true
                  nonEmpty current
    }

let rec push tree ps =
    match tree with
    | Leaf           -> ps
    | Node (l, a, r) -> push l ((a, r) :: ps)

let fromTree (tree : Tree<'elem>) : IEnumerator<'elem> =
    let mutable started  = false 
    let mutable pennants = push tree []
    { new IEnumerator<'elem> with
          member self.Current =
              if started then
                match pennants with
                | []           -> alreadyFinished ()
                | (a, t) :: ps -> a
              else 
                  notStarted ()
          member self.MoveNext () =
              if started then
                match pennants with
                | []           -> false
                | (a, t) :: ps -> pennants <- push t ps
                                  nonEmpty pennants
              else 
                  started <- true
                  nonEmpty pennants
    }

(*
> let e = fromList [] ;;
val e : IEnumerator<Int>
> e.Current ;; 
System.InvalidOperationException: Enumeration has not started. Call MoveNext.
> e.MoveNext () ;;
val it : Bool = false
> e.Current ;;                                   
System.InvalidOperationException: Enumeration already finished.

> let e = fromList [1];;                        
val e : IEnumerator<Int>
> e.Current ;;           
System.InvalidOperationException: Enumeration has not started. Call MoveNext.
> e.MoveNext () ;;       
val it : Bool = true
> e.Current ;;    
val it : Int = 1
> e.MoveNext () ;;
val it : Bool = false
> e.Current ;;    
System.InvalidOperationException: Enumeration already finished.
*)

let append (first : IEnumerator<'elem>) (second : IEnumerator<'elem>) =
    let mutable firstActive = true
    { new IEnumerator<'elem> with
          member self.Current =
              if firstActive then
                  first.Current
              else 
                  second.Current  
          member self.MoveNext () =
              if firstActive then
                  firstActive <- first.MoveNext ()
                  firstActive || second.MoveNext ()
              else 
                  second.MoveNext ()
    }

///////////////////////////////////////////////////////////////////////////////
// enumerables
///////////////////////////////////////////////////////////////////////////////

type IEnumerable<'elem> =
  interface
    abstract member GetEnumerator : Unit -> IEnumerator<'elem>
  end

let foreach (enumerable : IEnumerable<'elem>) (body : 'elem -> Unit) =
    let enumerator = enumerable.GetEnumerator()
    while enumerator.MoveNext () do
        body enumerator.Current

let Range (lower : Int, upper : Int) =
    { new IEnumerable<Int> with 
        member self.GetEnumerator() = range (lower, upper)
    }

let Append (first : IEnumerable<'elem>) (second : IEnumerable<'elem>) =
    { new IEnumerable<'elem> with
          member self.GetEnumerator () =
            append (first.GetEnumerator ()) (second.GetEnumerator ())
    }

let Wrap (xs : IEnumerator<'a>) =
    { new IEnumerable<'a> with 
        member self.GetEnumerator() = xs
    }

(*
let e = Append (Range (2, 3)) (Wrap (fromList [4711; 815])) ;;
foreach e print ;;
*)
