(*
Observer design pattern (once using abstract classes and subclasses
and a second time using higher-order functions).

fsharpi Mini.fs Objects/Observer.fs
open Objects.Observer ;;
*)

module Objects.Observer

open Mini

(*
type Observer<'state> =
    interface
        abstract Update : 'state -> Unit
    end
*)

type Observer<'state> =
    'state -> Unit

[<AbstractClass>]
type Subject<'state> () =
    class
        let mutable observers = []

        abstract member State : 'state with get, set
        
        member self.Attach (observer : Observer<'state>) =
            observers <- observer :: observers
            observer self.State   // notify observer

        member internal self.Notify () =
            for observer in observers do
                observer self.State
    end

let counter () =
      let mutable n = 0N
      {
        new Subject<Nat> () with
          member self.State
              with set k  = n <- k
                            self.Notify ()
              and  get () = n
      }

let steps = counter ()
do steps.Attach (fun n -> if n % 1000N = 1N then putchar '*')

let rec fibonacci (n : Nat) : Nat =
  steps.State <- steps.State + 1N
  if n <= 1N then n
             else fibonacci (n - 1N) + fibonacci (n - 2N)

// steps.State <- 0N ;;
// fibonacci 20N ;;
// steps.State ;;

module HOF =
  type IState<'state> =
      interface
          abstract member State : 'state with get, set
      end

  // simplified counter with exactly one observer
  let simpleCounter (observer : Nat -> Unit) =
        let mutable n = 0N
        {
          new IState<Nat> with
            member self.State
                with set k  = n <- k
                              observer n
                and  get () = n
        }

  let steps1 = simpleCounter (fun n -> if n % 1000N = 1N then putchar '*')

  type Observable<'state> (subject : (Unit -> Unit) -> IState<'state>) as self =
      class
          let state = subject self.Notify

          let mutable observers = []

          member self.Attach (observer : Observer<'state>) =
              observers <- observer :: observers
              observer state.State   // notify observer

          member private self.Notify () =
              for observer in observers do
                  observer state.State

          member self.State
            with get () = state.State
            and  set v  = state.State <- v

          interface IState<'state> with
            member self.State
              with get () = self.State
              and  set v  = self.State <- v
      end

  let counter (notify : Unit -> Unit) =
        let mutable n = 0N
        {
          new IState<Nat> with
            member self.State
                with set k  = n <- k
                              notify ()
                and  get () = n
        }

  let steps2 = Observable<Nat> counter
  do steps2.Attach (fun n -> if n % 1000N = 1N then putchar '*')

  let rec fibonacci (n : Nat) : Nat =
//    steps1.State <- steps1.State + 1N
    steps2.State <- steps2.State + 1N
    if n <= 1N then n
               else fibonacci (n - 1N) + fibonacci (n - 2N)

// HOF.steps2.State <- 0N ;;
// HOF.fibonacci 20N ;;
// HOF.steps2.State ;;
